"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getInstructionFunctionHighLevelFragment = void 0;
const shared_1 = require("../../../shared");
const ContextMap_1 = require("../ContextMap");
const asyncHelpers_1 = require("../asyncHelpers");
const common_1 = require("./common");
const instructionBytesCreatedOnChain_1 = require("./instructionBytesCreatedOnChain");
const instructionInputResolved_1 = require("./instructionInputResolved");
const instructionInputType_1 = require("./instructionInputType");
const instructionRemainingAccounts_1 = require("./instructionRemainingAccounts");
function getInstructionFunctionHighLevelFragment(scope) {
    const { useAsync, instructionNode, programNode, resolvedInputs, renamedArgs, dataArgsManifest, asyncResolvers, } = scope;
    if (useAsync && !(0, asyncHelpers_1.hasAsyncDefaultValues)(resolvedInputs, asyncResolvers)) {
        return (0, common_1.fragment)('');
    }
    const hasAccounts = instructionNode.accounts.length > 0;
    const hasDataArgs = !!instructionNode.dataArgs.link ||
        instructionNode.dataArgs.struct.fields.filter((field) => { var _a; return ((_a = field.defaultsTo) === null || _a === void 0 ? void 0 : _a.strategy) !== 'omitted'; }).length > 0;
    const hasExtraArgs = !!instructionNode.extraArgs.link ||
        instructionNode.extraArgs.struct.fields.filter((field) => { var _a; return ((_a = field.defaultsTo) === null || _a === void 0 ? void 0 : _a.strategy) !== 'omitted'; }).length > 0;
    const hasAnyArgs = hasDataArgs || hasExtraArgs;
    const argsTypeFragment = (0, common_1.fragment)(instructionNode.dataArgs.link
        ? dataArgsManifest.looseType.render
        : `${(0, shared_1.pascalCase)(instructionNode.dataArgs.name)}Args`);
    if (instructionNode.dataArgs.link) {
        argsTypeFragment.mergeImportsWith(dataArgsManifest.looseType);
    }
    const functionName = `get${(0, shared_1.pascalCase)(instructionNode.name)}Instruction${useAsync ? 'Async' : ''}`;
    const lowLevelFunctionName = `get${(0, shared_1.pascalCase)(instructionNode.name)}InstructionRaw`;
    const typeParamsFragment = getTypeParams(instructionNode, programNode);
    const instructionTypeFragment = getInstructionType(instructionNode, false);
    const instructionTypeWithSignersFragment = getInstructionType(instructionNode, true);
    const wrapInPromiseIfAsync = (value) => useAsync ? `Promise<${value}>` : value;
    // Input.
    const inputTypeFragment = (0, instructionInputType_1.getInstructionInputTypeFragment)({
        ...scope,
        withSigners: false,
    });
    const inputTypeWithSignersFragment = (0, instructionInputType_1.getInstructionInputTypeFragment)({
        ...scope,
        withSigners: true,
    });
    const inputTypeCallFragment = getInputTypeCall({
        ...scope,
        withSigners: false,
    });
    const inputTypeCallWithSignersFragment = getInputTypeCall({
        ...scope,
        withSigners: true,
    });
    const renamedArgsText = [...renamedArgs.entries()]
        .map(([k, v]) => `${k}: input.${v}`)
        .join(', ');
    const resolvedInputsFragment = (0, instructionInputResolved_1.getInstructionInputResolvedFragment)(scope);
    const remainingAccountsFragment = (0, instructionRemainingAccounts_1.getInstructionRemainingAccountsFragment)(scope);
    const bytesCreatedOnChainFragment = (0, instructionBytesCreatedOnChain_1.getInstructionBytesCreatedOnChainFragment)(scope);
    const context = new ContextMap_1.ContextMap()
        .add('getProgramAddress')
        .mergeWith(resolvedInputsFragment.interfaces, remainingAccountsFragment.interfaces, bytesCreatedOnChainFragment.interfaces);
    const contextFragment = context.toFragment();
    const functionFragment = (0, common_1.fragmentFromTemplate)('instructionFunctionHighLevel.njk', {
        instruction: instructionNode,
        program: programNode,
        hasAccounts,
        hasDataArgs,
        hasExtraArgs,
        hasAnyArgs,
        argsTypeFragment,
        functionName,
        lowLevelFunctionName,
        typeParamsFragment,
        instructionTypeFragment,
        instructionTypeWithSignersFragment,
        inputTypeFragment,
        inputTypeWithSignersFragment,
        inputTypeCallFragment,
        inputTypeCallWithSignersFragment,
        contextFragment,
        renamedArgs: renamedArgsText,
        resolvedInputsFragment,
        remainingAccountsFragment,
        bytesCreatedOnChainFragment,
        useAsync,
        wrapInPromiseIfAsync,
    })
        .mergeImportsWith(typeParamsFragment, instructionTypeFragment, instructionTypeWithSignersFragment, inputTypeFragment, inputTypeWithSignersFragment, inputTypeCallFragment, inputTypeCallWithSignersFragment, contextFragment, resolvedInputsFragment, remainingAccountsFragment, bytesCreatedOnChainFragment, argsTypeFragment)
        .addImports('solanaAddresses', ['Address']);
    if (hasAccounts) {
        functionFragment
            .addImports('solanaInstructions', ['IAccountMeta'])
            .addImports('shared', ['getAccountMetasWithSigners', 'ResolvedAccount']);
    }
    return functionFragment;
}
exports.getInstructionFunctionHighLevelFragment = getInstructionFunctionHighLevelFragment;
function getTypeParams(instructionNode, programNode) {
    const typeParams = [
        ...instructionNode.accounts.map((account) => `TAccount${(0, shared_1.pascalCase)(account.name)} extends string`),
        `TProgram extends string = "${programNode.publicKey}"`,
    ];
    return (0, common_1.fragment)(typeParams.filter((x) => !!x).join(', '));
}
function getInstructionType(instructionNode, withSigners) {
    const instructionTypeName = (0, shared_1.pascalCase)(`${instructionNode.name}Instruction${withSigners ? 'WithSigners' : ''}`);
    const accountTypeParamsFragments = instructionNode.accounts.map((account) => {
        const typeParam = `TAccount${(0, shared_1.pascalCase)(account.name)}`;
        const camelName = (0, shared_1.camelCase)(account.name);
        if (account.isSigner === 'either' && withSigners) {
            const signerRole = account.isWritable
                ? 'WritableSignerAccount'
                : 'ReadonlySignerAccount';
            return (0, common_1.fragment)(`typeof input["${camelName}"] extends TransactionSigner<${typeParam}> ` +
                `? ${signerRole}<${typeParam}> & IAccountSignerMeta<${typeParam}> ` +
                `: ${typeParam}`)
                .addImports('solanaInstructions', [signerRole])
                .addImports('solanaSigners', ['IAccountSignerMeta']);
        }
        return (0, common_1.fragment)(typeParam);
    });
    return (0, common_1.mergeFragments)([(0, common_1.fragment)('TProgram'), ...accountTypeParamsFragments], (renders) => renders.join(', ')).mapRender((r) => `${instructionTypeName}<${r}>`);
}
function getInputTypeCall(scope) {
    const { instructionNode, withSigners, useAsync } = scope;
    const inputTypeName = (0, shared_1.pascalCase)(`${instructionNode.name}${useAsync ? 'Async' : ''}Input${withSigners ? 'WithSigners' : ''}`);
    if (instructionNode.accounts.length === 0)
        return (0, common_1.fragment)(inputTypeName);
    const accountTypeParams = instructionNode.accounts
        .map((account) => `TAccount${(0, shared_1.pascalCase)(account.name)}`)
        .join(', ');
    return (0, common_1.fragment)(`${inputTypeName}<${accountTypeParams}>`);
}
//# sourceMappingURL=instructionFunctionHighLevel.js.map